#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <poll.h>

#include "socket.h"


//////////////////////////////////////////////////////////////////////////////
// Constructors...
//

//
// Copy-constructor
//
Socket::Socket(const Socket& s)
{
	sfd   = s.sfd;
	sport = s.sport;
}

//
// Create a Socket from a sock-id
//
Socket::Socket(int fd)
{
	sfd = fd;
	sport = 0;
	if (fd > 0) {
		/* get the "real" socket information ... */
		sockaddr_in sa;
		socklen_t ssize = sizeof(sa);
		if (getsockname(fd, (sockaddr*) &sa, &ssize))
			return;
		sport = ntohs(sa.sin_port);
	}
}


//////////////////////////////////////////////////////////////////////////////
// Destructor...
//
Socket::~Socket() {
	Close();
}


//////////////////////////////////////////////////////////////////////////////
// Create a new socket, at the specified port.
// If @port is zero, then the OS will give you one at random.
// Return true on success, false on failure.
//
bool Socket::Open(const char *proto_name, unsigned short port)
{
	if (sfd >= 0)
		return false;
	if (!proto_name)
		return false;

	protoent *my_proto = (protoent*) getprotobyname(proto_name);
	if (!my_proto)
		return false;

	switch (my_proto->p_proto) {
	    case IPPROTO_UDP:
		sfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
		break;
	    case IPPROTO_TCP:
		sfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		break;
	    default:
		return false;
	}
	if (sfd < 0)
		return false;

	sockaddr_in sa;
	sa.sin_addr.s_addr = INADDR_ANY;
	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	memset(&sa.sin_zero, 0, 8);

	if (bind(sfd, (sockaddr*) &sa, sizeof(sa))) {
		close(sfd);
		return false;
	}

	/* get the "real" socket information ... */
	sport = port;
	socklen_t ssize = sizeof(sa);
	if (getsockname(sfd, (sockaddr*) &sa, &ssize))
		return true;
	sport = ntohs(sa.sin_port);
	return true;
}


//////////////////////////////////////////////////////////////////////////////
// Destroy this socket and close all connections.
// Return true on success, false on failure
//
bool Socket::Close(void)
{
	/* try to close the socket if it's open */
	if ((sfd >= 0) && close(sfd))
		return false;

	sfd = -1;
	sport = 0;
	return true;
}


//////////////////////////////////////////////////////////////////////////////
// Enable connections for this socket
//
bool Socket::Listen(int backlog) {
	return (sfd >= 0) && !listen(sfd, backlog);
}


//////////////////////////////////////////////////////////////////////////////
// Accept a connection from a client.
// It returns a new Socket on success, 0 on failure.
// IMPORTANT: This only works for server sockets!
//
Socket* Socket::Accept(void)
{
	if (sfd < 0)
		return false;

	sockaddr_in sa;
	socklen_t ssize = sizeof(sa);
	int fd = accept(sfd, (sockaddr*) &sa, &ssize);
	if (fd < 0)
		return (Socket*) 0;

	Socket *p = new Socket(fd);
	if (!p)
		close(fd);

	return p;
}


//////////////////////////////////////////////////////////////////////////////
// Connect to another host.
//	 @host can be either an IP address or a hostname string.
//		if @host is NULL then we assume we want to connect to
//		the local host
//	 @port is the port number (duh!)
//
// Return true on succes, false on failure
//
bool Socket::Connect(const char *host, unsigned short port)
{
	if (sfd < 0)
		return false;

	sockaddr_in  sa;
	unsigned long addr;

	/* conver the @host to a network address */
	if (!host) {
		addr = INADDR_ANY;
	} else if (isdigit(host[0])) {
		// try IP string first
		addr = inet_addr(host);
		if (addr < 0)
			return false;
	} else {
		// try HOSTNAME string second
		hostent *sh = gethostbyname(host);
		if (!sh)
			return false;
		addr = ((unsigned*) sh->h_addr)[0];
	}

	/* construct a sockaddr_in struct for the connect() syscall */
	sa.sin_addr.s_addr = addr;
	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	memset(&sa.sin_zero, 0, 8);

	if (connect(sfd, (sockaddr*) &sa, sizeof(sa)))
		return false;

	/* get the "real" socket information ... */
	sport = port;
	socklen_t ssize = sizeof(sa);
	if (getsockname(sfd, (sockaddr*) &sa, &ssize))
		return true;
	sport = ntohs(sa.sin_port);
	return true;
}


//////////////////////////////////////////////////////////////////////////////
// Block and wait for the socket to have data to receive.
// It uses the poll() syscall.
// If @msec is -1 it will wait for ever.
// If @msec is 0 it will return without blocking.
// see the poll man page for more info
//
// Return true if there is some data to be read, false otherwise
// (ie. if the timeout expired and there is no data)
//
bool Socket::hasData(int msec)
{
	if (sfd < 0)
		return false;

	pollfd  fds[1];
	nfds_t  nfds = 1;

	fds[0].fd = sfd;
	fds[0].events = POLLIN;
	fds[0].revents = 0;

	int retv = poll(fds, nfds, msec);
	if (retv <= 0)
		return false;

	return (fds[0].revents & POLLIN) != 0;
}


//////////////////////////////////////////////////////////////////////////////
// Send a buffer over the network
// Return the number of bytes actually sent.
// Return -1 on error.
//

//
// connected send
//
int Socket::Send(const void* buff, int size)
{
	if ((sfd < 0) || !buff || (size < 0))
		return -1;

	char *ptr = (char*) buff;
	while (size > 0) {
		int num = send(sfd, ptr, size, 0);
		if (num < 0)
			return -1;
		size -= num;
		ptr += num;
	}
	return ptr - (char*) buff;
}

//
// disconnected send
//
int Socket::SendTo(struct sockaddr_in *to, const void* buff, int size)
{
	if ((sfd < 0) || !to || !buff || (size < 0))
		return -1;

	char *ptr = (char*) buff;
	while (size > 0) {
		int num = sendto(sfd, ptr, size, 0,
				 (sockaddr*) to, sizeof(*to));
		if (num < 0)
			return -1;
		size -= num;
		ptr += num;
	}
	return ptr - (char*) buff;
}

//
// also disconnected send
//
int Socket::SendTo(const char* host, unsigned short port,
		   const void* buff, int size)
{
	if ((sfd < 0) || !port || !buff || (size < 0))
		return -1;

	sockaddr_in  sa;
	unsigned long addr;

	/* conver the @host to a network address */
	if (!host) {
		addr = INADDR_ANY;
	} else if (isdigit(host[0])) {
		// try IP string first
		addr = inet_addr(host);
		if (addr < 0)
			return false;
	} else {
		// try HOSTNAME string second
		hostent *sh = gethostbyname(host);
		if (!sh)
			return false;
		addr = ((unsigned*) sh->h_addr)[0];
	}

	/* construct a sockaddr_in struct */
	sa.sin_addr.s_addr = addr;
	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	memset(&sa.sin_zero, 0, 8);

	return SendTo(&sa, buff, size);
}


//////////////////////////////////////////////////////////////////////////////
// Receive a buffer over the network
// Return the actual number of bytes received.
// Return -1 on error.
//

// 
// connected receive
//
int Socket::Recv(void* buff, int maxlen)
{
	if ((sfd < 0) || !buff || (maxlen <= 0))
		return -1;
	return recv(sfd, buff, maxlen, 0);
}

//
// disconnected receive
//
int Socket::RecvFrom(struct sockaddr_in *frm, void* buff, int maxlen)
{
	if ((sfd < 0) || !buff || (maxlen <= 0))
		return -1;
	
	socklen_t frmlen;
	return recvfrom(sfd, buff, maxlen, 0, (sockaddr*) frm, &frmlen);
}

